@@ -41,29 +41,76 @@
 #include <glib.h>
 
 #include <openobex/obex.h>
-#include <openobex/obex_const.h>
 
 #include "logging.h"
 #include "obex.h"
+#include "obex-priv.h"
+#include "server.h"
 #include "dbus.h"
 #include "mimetype.h"
 #include "service.h"
+#include "transport.h"
+#include "btio.h"
 
 /* Default MTU's */
 #define DEFAULT_RX_MTU 32767
 #define DEFAULT_TX_MTU 32767
 
 /* Connection ID */
-static guint32 cid = 0x0000;
+static uint32_t cid = 0x0000;
 
 static GSList *sessions = NULL;
 
 typedef struct {
-	guint8  version;
-	guint8  flags;
-	guint16 mtu;
+	uint8_t  version;
+	uint8_t  flags;
+	uint16_t mtu;
 } __attribute__ ((packed)) obex_connect_hdr_t;
 
+static void os_set_response(obex_object_t *obj, int err)
+{
+	uint8_t rsp;
+	uint8_t lastrsp;
+
+	switch (err) {
+	case 0:
+		rsp = OBEX_RSP_CONTINUE;
+		lastrsp = OBEX_RSP_SUCCESS;
+		break;
+	case -EPERM:
+	case -EACCES:
+		rsp = OBEX_RSP_FORBIDDEN;
+		lastrsp = OBEX_RSP_FORBIDDEN;
+		break;
+	case -ENOENT:
+		rsp = OBEX_RSP_NOT_FOUND;
+		lastrsp = OBEX_RSP_NOT_FOUND;
+		break;
+	case -EBADR:
+		rsp = OBEX_RSP_BAD_REQUEST;
+		lastrsp = OBEX_RSP_BAD_REQUEST;
+		break;
+	case -EFAULT:
+		rsp = OBEX_RSP_SERVICE_UNAVAILABLE;
+		lastrsp = OBEX_RSP_SERVICE_UNAVAILABLE;
+		break;
+	case -EINVAL:
+		rsp = OBEX_RSP_NOT_IMPLEMENTED;
+		lastrsp = OBEX_RSP_NOT_IMPLEMENTED;
+		break;
+	case -ENOTEMPTY:
+	case -EEXIST:
+		rsp = OBEX_RSP_PRECONDITION_FAILED;
+		lastrsp = OBEX_RSP_PRECONDITION_FAILED;
+		break;
+	default:
+		rsp = OBEX_RSP_INTERNAL_SERVER_ERROR;
+		lastrsp = OBEX_RSP_INTERNAL_SERVER_ERROR;
+	}
+
+	OBEX_ObjectSetRsp(obj, rsp, lastrsp);
+}
+
 static void os_reset_session(struct obex_session *os)
 {
 	if (os->object) {
@@ -71,12 +118,8 @@ static void os_reset_session(struct obex_session *os)
 		os->driver->close(os->object);
 		os->object = NULL;
 		os->obj = NULL;
-		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->current_folder) {
-			gchar *path;
-			path = g_build_filename(os->current_folder, os->name, NULL);
-			os->driver->remove(path);
-			g_free(path);
-		}
+		if (os->aborted && os->cmd == OBEX_CMD_PUT && os->path)
+			os->driver->remove(os->path);
 	}
 
 	if (os->name) {
@@ -91,9 +134,14 @@ static void os_reset_session(struct obex_session *os)
 		g_free(os->buf);
 		os->buf = NULL;
 	}
+	if (os->path) {
+		g_free(os->path);
+		os->path = NULL;
+	}
 
 	os->driver = NULL;
 	os->aborted = FALSE;
+	os->pending = 0;
 	os->offset = 0;
 	os->size = OBJECT_SIZE_DELETE;
 	os->finished = 0;
@@ -115,9 +163,6 @@ static void obex_session_free(struct obex_session *os)
 
 	os_reset_session(os);
 
-	if (os->current_folder)
-		g_free(os->current_folder);
-
 	if (os->io)
 		g_io_channel_unref(os->io);
 
@@ -125,12 +170,12 @@ static void obex_session_free(struct obex_session *os)
 }
 
 /* From Imendio's GnomeVFS OBEX module (om-utils.c) */
-static time_t parse_iso8610(const gchar *val, int size)
+static time_t parse_iso8610(const char *val, int size)
 {
 	time_t time, tz_offset = 0;
 	struct tm tm;
-	gchar *date;
-	gchar tz;
+	char *date;
+	char tz;
 	int nr;
 
 	memset(&tm, 0, sizeof(tm));
@@ -177,11 +222,12 @@ static void cmd_connect(struct obex_session *os,
 	obex_connect_hdr_t *nonhdr;
 	obex_headerdata_t hd;
 	uint8_t *buffer;
-	guint hlen, newsize;
-	guint16 mtu;
-	guint8 hi;
-	const guint8 *target = NULL, *who = NULL;
-	guint target_size = 0, who_size = 0;
+	unsigned int hlen, newsize;
+	uint16_t mtu;
+	uint8_t hi;
+	const uint8_t *target = NULL, *who = NULL;
+	unsigned int target_size = 0, who_size = 0;
+	int err;
 
 	if (OBEX_ObjectGetNonHdrData(obj, &buffer) != sizeof(*nonhdr)) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -228,16 +274,32 @@ static void cmd_connect(struct obex_session *os,
 
 	debug("Selected driver: %s", os->service->name);
 
-	if (os->service->connect)
-		os->service->connect(obex, obj);
+	if (!os->service->connect) {
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+		return;
+	}
+
+	os->service_data = os->service->connect(os, &err);
+	if (err == 0 && os->service->target) {
+		hd.bs = os->service->target;
+		OBEX_ObjectAddHeader(obex, obj,
+				OBEX_HDR_WHO, hd, 16,
+				OBEX_FL_FIT_ONE_PACKET);
+		hd.bq4 = os->cid;
+		OBEX_ObjectAddHeader(obex, obj,
+				OBEX_HDR_CONNECTION, hd, 4,
+				OBEX_FL_FIT_ONE_PACKET);
+	}
+
+	os_set_response(obj, err);
 }
 
-static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
+static gboolean chk_cid(obex_t *obex, obex_object_t *obj, uint32_t cid)
 {
 	struct obex_session *os;
 	obex_headerdata_t hd;
-	guint hlen;
-	guint8 hi;
+	unsigned int hlen;
+	uint8_t hi;
 	gboolean ret = FALSE;
 
 	os = OBEX_GetUserData(obex);
@@ -262,11 +324,182 @@ static gboolean chk_cid(obex_t *obex, obex_object_t *obj, guint32 cid)
 	return ret;
 }
 
+static int obex_read_stream(struct obex_session *os, obex_t *obex,
+						obex_object_t *obj)
+{
+	int size;
+	int32_t len = 0;
+	const uint8_t *buffer;
+
+	if (os->aborted)
+		return -EPERM;
+
+	/* workaround: client didn't send the object lenght */
+	if (os->size == OBJECT_SIZE_DELETE)
+		os->size = OBJECT_SIZE_UNKNOWN;
+
+	/* If there's something to write and we are able to write it */
+	if (os->pending > 0 && os->driver)
+		goto write;
+
+	size = OBEX_ObjectReadStream(obex, obj, &buffer);
+	if (size < 0) {
+		error("Error on OBEX stream");
+		return -EIO;
+	}
+
+	if (size > os->rx_mtu) {
+		error("Received more data than RX_MAX");
+		return -EIO;
+	}
+
+	os->buf = g_realloc(os->buf, os->pending + size);
+	memcpy(os->buf + os->pending, buffer, size);
+	os->pending += size;
+	if (os->object == NULL) {
+		debug("Stored %u bytes into temporary buffer", os->pending);
+		return 0;
+	}
+
+write:
+	while (os->pending > 0) {
+		int w;
+
+		w = os->driver->write(os->object, os->buf + len,
+					os->pending);
+		if (w < 0) {
+			if (w == -EINTR)
+				continue;
+			else {
+				memmove(os->buf, os->buf + len, os->pending);
+				return w;
+			}
+		}
+
+		len += w;
+		os->offset += w;
+		os->pending -= w;
+	}
+
+	return 0;
+}
+
+static int obex_write_stream(struct obex_session *os,
+			obex_t *obex, obex_object_t *obj)
+{
+	obex_headerdata_t hd;
+	uint8_t *ptr;
+	int32_t len;
+	unsigned int flags;
+	uint8_t hi;
+
+	debug("obex_write_stream: name=%s type=%s tx_mtu=%d file=%p",
+		os->name ? os->name : "", os->type ? os->type : "",
+		os->tx_mtu, os->object);
+
+	if (os->aborted)
+		return -EPERM;
+
+	if (os->object == NULL) {
+		if (os->buf == NULL && os->finished == FALSE)
+			return -EIO;
+
+		len = MIN(os->size - os->offset, os->tx_mtu);
+		ptr = os->buf + os->offset;
+		goto add_header;
+	}
+
+	len = os->driver->read(os->object, os->buf, os->tx_mtu, &hi);
+	if (len < 0) {
+		error("read(): %s (%d)", strerror(-len), -len);
+		if (len == -EAGAIN)
+			return len;
+		else if (len == -ENOSTR)
+			return 0;
+
+		g_free(os->buf);
+		os->buf = NULL;
+		return len;
+	}
+
+	ptr = os->buf;
+
+add_header:
+
+	hd.bs = ptr;
+
+	switch (hi) {
+	case OBEX_HDR_BODY:
+		flags = len ? OBEX_FL_STREAM_DATA : OBEX_FL_STREAM_DATAEND;
+		break;
+	case OBEX_HDR_APPARAM:
+		flags =  0;
+		break;
+	}
+
+	OBEX_ObjectAddHeader(obex, obj, hi, hd, len, flags);
+
+	if (len == 0) {
+		g_free(os->buf);
+		os->buf = NULL;
+		return len;
+	}
+
+	os->offset += len;
+
+	return len;
+}
+
+static gboolean handle_async_io(void *object, int flags, int err,
+						void *user_data)
+{
+	struct obex_session *os = user_data;
+	int ret = 0;
+
+	if (err < 0) {
+		ret = err;
+		goto proceed;
+	}
+
+	if (flags & (G_IO_IN | G_IO_PRI))
+		ret = obex_write_stream(os, os->obex, os->obj);
+	else if (flags & G_IO_OUT)
+		ret = obex_read_stream(os, os->obex, os->obj);
+
+proceed:
+	switch (ret) {
+	case -EINVAL:
+		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_BAD_REQUEST,
+				OBEX_RSP_BAD_REQUEST);
+		break;
+	case -EPERM:
+		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_FORBIDDEN,
+					OBEX_RSP_FORBIDDEN);
+		break;
+	case -ENOENT:
+		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_NOT_FOUND,
+							OBEX_RSP_NOT_FOUND);
+		break;
+	default:
+		if (ret < 0)
+			OBEX_ObjectSetRsp(os->obj,
+					OBEX_RSP_INTERNAL_SERVER_ERROR,
+					OBEX_RSP_INTERNAL_SERVER_ERROR);
+		break;
+	}
+
+	OBEX_ResumeRequest(os->obex);
+
+	return FALSE;
+}
+
 static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 {
 	obex_headerdata_t hd;
-	guint hlen;
-	guint8 hi;
+	gboolean stream;
+	unsigned int hlen;
+	uint8_t hi;
+	int err;
 
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -290,7 +523,7 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			if (hlen == 0)
 				continue;
 
-			os->name = g_convert((const gchar *) hd.bs, hlen,
+			os->name = g_convert((const char *) hd.bs, hlen,
 					"UTF8", "UTF16BE", NULL, NULL, NULL);
 			debug("OBEX_HDR_NAME: %s", os->name);
 			break;
@@ -307,39 +540,74 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			if (hd.bs[hlen - 1] != '\0')
 				break;
 
-			if (!g_utf8_validate((const gchar *) hd.bs, -1, NULL)) {
+			if (!g_utf8_validate((const char *) hd.bs, -1, NULL)) {
 				debug("Invalid type header: %s", hd.bs);
 				break;
 			}
 
 			/* FIXME: x-obex/folder-listing - type is mandatory */
 
-			os->type = g_strndup((const gchar *) hd.bs, hlen);
+			os->type = g_strndup((const char *) hd.bs, hlen);
 			debug("OBEX_HDR_TYPE: %s", os->type);
-			os->driver = obex_mime_type_driver_find(os->service->target, os->type);
+			os->driver = obex_mime_type_driver_find(
+						os->service->target, os->type,
+						os->service->who,
+						os->service->who_size);
 			break;
 		}
 	}
 
+	if (os->type == NULL)
+		os->driver = obex_mime_type_driver_find(os->service->target,
+							NULL,
+							os->service->who,
+							os->service->who_size);
+
 	if (!os->driver) {
-		os->driver = obex_mime_type_driver_find(os->service->target, NULL);
-		if (!os->driver) {
-			error("No driver found");
-			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+		error("No driver found");
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 					OBEX_RSP_NOT_IMPLEMENTED);
-			return;
-		}
+		return;
 	}
 
-	os->service->get(obex, obj);
+	err = os->service->get(os, obj, &stream, os->service_data);
+
+	if (err < 0)
+		goto done;
+
+	if (os->size != OBJECT_SIZE_UNKNOWN) {
+		hd.bq4 = os->size;
+		OBEX_ObjectAddHeader(obex, obj,
+				OBEX_HDR_LENGTH, hd, 4, 0);
+	}
+
+	/* Add body header */
+	hd.bs = NULL;
+	if (os->size == 0)
+		OBEX_ObjectAddHeader (obex, obj, OBEX_HDR_BODY,
+				hd, 0, OBEX_FL_FIT_ONE_PACKET);
+	else if (!stream) {
+		/* Asynchronous operation that doesn't use stream */
+		OBEX_SuspendRequest(obex, obj);
+		os->obj = obj;
+		os->driver->set_io_watch(os->object, handle_async_io, os);
+		return;
+	} else
+		/* Standard data stream */
+		OBEX_ObjectAddHeader (obex, obj, OBEX_HDR_BODY,
+				hd, 0, OBEX_FL_STREAM_START);
+
+done:
+	os_set_response(obj, err);
 }
 
 static void cmd_setpath(struct obex_session *os,
 			obex_t *obex, obex_object_t *obj)
 {
 	obex_headerdata_t hd;
-	guint32 hlen;
-	guint8 hi;
+	uint32_t hlen;
+	int err;
+	uint8_t hi;
 
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
@@ -373,22 +641,25 @@ static void cmd_setpath(struct obex_session *os,
 			break;
 		}
 
-		os->name = g_convert((const gchar *) hd.bs, hlen,
+		os->name = g_convert((const char *) hd.bs, hlen,
 				"UTF8", "UTF16BE", NULL, NULL, NULL);
 
 		debug("Set path name: %s", os->name);
 		break;
 	}
 
-	os->service->setpath(obex, obj);
+	err = os->service->setpath(os, obj, os->service_data);
+	os_set_response(obj, err);
 }
 
-int os_prepare_get(struct obex_session *os, gchar *filename, size_t *size)
+int obex_get_stream_start(struct obex_session *os, const char *filename)
 {
-	gint err;
-	gpointer object;
+	int err;
+	void *object;
+	size_t size;
 
-	object = os->driver->open(filename, O_RDONLY, 0, size, &err);
+	object = os->driver->open(filename, O_RDONLY, 0, os->service_data,
+								&size, &err);
 	if (object == NULL) {
 		error("open(%s): %s (%d)", filename, strerror(-err), -err);
 		goto fail;
@@ -396,8 +667,9 @@ int os_prepare_get(struct obex_session *os, gchar *filename, size_t *size)
 
 	os->object = object;
 	os->offset = 0;
+	os->size = size;
 
-	if (*size > 0)
+	if (size > 0)
 		os->buf = g_malloc0(os->tx_mtu);
 
 	return 0;
@@ -405,205 +677,42 @@ int os_prepare_get(struct obex_session *os, gchar *filename, size_t *size)
 fail:
 	if (object)
 		os->driver->close(object);
+
 	return err;
 }
 
-static gint obex_write_stream(struct obex_session *os,
-			obex_t *obex, obex_object_t *obj)
+int obex_put_stream_start(struct obex_session *os, const char *filename)
 {
-	obex_headerdata_t hd;
-	gint32 len;
-	guint8 *ptr;
-
-	debug("obex_write_stream: name=%s type=%s tx_mtu=%d file=%p",
-		os->name ? os->name : "", os->type ? os->type : "",
-		os->tx_mtu, os->object);
-
-	if (os->aborted)
-		return -EPERM;
-
-	if (os->object == NULL) {
-		if (os->buf == NULL && os->finished == FALSE)
-			return -EIO;
-
-		len = MIN(os->size - os->offset, os->tx_mtu);
-		ptr = os->buf + os->offset;
-		goto add_header;
-	}
-
-	len = os->driver->read(os->object, os->buf, os->tx_mtu);
-	if (len < 0) {
-		error("read(): %s (%d)", strerror(-len), -len);
-		if (len == -EAGAIN)
-			return len;
-
-		g_free(os->buf);
-		os->buf = NULL;
-		return len;
-	}
-
-	ptr = os->buf;
-
-add_header:
-
-	hd.bs = ptr;
-
-	if (len == 0) {
-		OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, 0,
-					OBEX_FL_STREAM_DATAEND);
-		g_free(os->buf);
-		os->buf = NULL;
-		return len;
-	}
-
-	os->offset += len;
-
-	OBEX_ObjectAddHeader(obex, obj, OBEX_HDR_BODY, hd, len,
-				OBEX_FL_STREAM_DATA);
-
-	return len;
-}
-
-gint os_prepare_put(struct obex_session *os)
-{
-	gchar *path;
-	gint len;
 	int err;
 
-	path = g_build_filename(os->current_folder, os->name, NULL);
-	os->object = os->driver->open(path, O_WRONLY | O_CREAT | O_TRUNC, 0600,
+	os->object = os->driver->open(filename, O_WRONLY | O_CREAT | O_TRUNC,
+					0600, os->service_data,
 					os->size != OBJECT_SIZE_UNKNOWN ?
 					(size_t *) &os->size : NULL, &err);
 	if (os->object == NULL) {
-		error("open(%s): %s (%d)", path, strerror(-err), -err);
-		g_free(path);
+		error("open(%s): %s (%d)", filename, strerror(-err), -err);
 		return -EPERM;
 	}
 
-	g_free(path);
+	os->path = g_strdup(filename);
 
 	if (!os->buf) {
 		debug("PUT request checked, no buffered data");
 		return 0;
 	}
 
-	len = 0;
-	while (len < os->offset) {
-		gint w;
-
-		w = os->driver->write(os->object, os->buf + len, os->offset - len);
-		if (w < 0) {
-			error("write(%s): %s (%d)", path, strerror(-w), -w);
-			if (w == -EINTR)
-				continue;
-			else
-				return w;
-		}
-
-		len += w;
-	}
-
-	return 0;
-}
-
-static gint obex_read_stream(struct obex_session *os, obex_t *obex,
-				obex_object_t *obj)
-{
-	gint size;
-	gint32 len = 0;
-	const guint8 *buffer;
-
-	if (os->aborted)
-		return -EPERM;
-
-	/* workaround: client didn't send the object lenght */
-	if (os->size == OBJECT_SIZE_DELETE)
-		os->size = OBJECT_SIZE_UNKNOWN;
-
-	size = OBEX_ObjectReadStream(obex, obj, &buffer);
-	if (size < 0) {
-		error("Error on OBEX stream");
-		return -EIO;
-	}
-
-	if (size > os->rx_mtu) {
-		error("Received more data than RX_MAX");
-		return -EIO;
-	}
-
-	if (os->object == NULL && size > 0) {
-		os->buf = g_realloc(os->buf, os->offset + size);
-		memcpy(os->buf + os->offset, buffer, size);
-		os->offset += size;
-
-		debug("Stored %u bytes into temporary buffer", size);
-
+	if (os->pending == 0)
 		return 0;
-	}
 
-	while (len < size) {
-		gint w;
-
-		w = os->driver->write(os->object, buffer + len, size - len);
-		if (w < 0) {
-			if (w == -EINTR)
-				continue;
-			else
-				return w;
-		}
-
-		len += w;
-	}
-
-	os->offset += len;
-
-	return 0;
-}
-
-static gboolean handle_async_io(gpointer object, int flags, int err,
-				gpointer user_data)
-{
-	struct obex_session *os = user_data;
-	int ret = 0;
-
-	if (err < 0) {
-		ret = err;
-		goto proceed;
-	}
-
-	if (flags & (G_IO_IN | G_IO_PRI))
-		ret = obex_write_stream(os, os->obex, os->obj);
-	else if (flags & G_IO_OUT)
-		ret = obex_read_stream(os, os->obex, os->obj);
-
-proceed:
-	switch (ret) {
-	case -EINVAL:
-		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_BAD_REQUEST,
-				OBEX_RSP_BAD_REQUEST);
-	case -EPERM:
-		OBEX_ObjectSetRsp(os->obj, OBEX_RSP_FORBIDDEN,
-					OBEX_RSP_FORBIDDEN);
-		break;
-	default:
-		if (ret < 0)
-			OBEX_ObjectSetRsp(os->obj,
-					OBEX_RSP_INTERNAL_SERVER_ERROR,
-					OBEX_RSP_INTERNAL_SERVER_ERROR);
-		break;
-	}
-
-	OBEX_ResumeRequest(os->obex);
-
-	return FALSE;
+	return obex_read_stream(os, os->obex, NULL);
 }
 
 static gboolean check_put(obex_t *obex, obex_object_t *obj)
 {
 	struct obex_session *os;
 	obex_headerdata_t hd;
-	guint hlen;
-	guint8 hi;
+	unsigned int hlen;
+	uint8_t hi;
 	int ret;
 
 	os = OBEX_GetUserData(obex);
@@ -629,7 +738,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 			if (hlen == 0)
 				continue;
 
-			os->name = g_convert((const gchar *) hd.bs, hlen,
+			os->name = g_convert((const char *) hd.bs, hlen,
 					"UTF8", "UTF16BE", NULL, NULL, NULL);
 			debug("OBEX_HDR_NAME: %s", os->name);
 			break;
@@ -647,14 +756,17 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 			if (hd.bs[hlen - 1] != '\0')
 				break;
 
-			if (!g_utf8_validate((const gchar *) hd.bs, -1, NULL)) {
+			if (!g_utf8_validate((const char *) hd.bs, -1, NULL)) {
 				debug("Invalid type header: %s", hd.bs);
 				break;
 			}
 
-			os->type = g_strndup((const gchar *) hd.bs, hlen);
+			os->type = g_strndup((const char *) hd.bs, hlen);
 			debug("OBEX_HDR_TYPE: %s", os->type);
-			os->driver = obex_mime_type_driver_find(os->service->target, os->type);
+			os->driver = obex_mime_type_driver_find(
+						os->service->target, os->type,
+						os->service->who,
+						os->service->who_size);
 			break;
 
 		case OBEX_HDR_BODY:
@@ -667,37 +779,40 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 			debug("OBEX_HDR_LENGTH: %d", os->size);
 			break;
 		case OBEX_HDR_TIME:
-			os->time = parse_iso8610((const gchar *) hd.bs, hlen);
+			os->time = parse_iso8610((const char *) hd.bs, hlen);
 			break;
 		}
 	}
 
 	OBEX_ObjectReParseHeaders(obex, obj);
 
+	if (os->type == NULL)
+		os->driver = obex_mime_type_driver_find(os->service->target,
+							NULL,
+							os->service->who,
+							os->service->who_size);
+
 	if (!os->driver) {
-		os->driver = obex_mime_type_driver_find(os->service->target, NULL);
-		if (!os->driver) {
-			error("No driver found");
-			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+		error("No driver found");
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
 					OBEX_RSP_NOT_IMPLEMENTED);
-			return FALSE;
-		}
+		return FALSE;
 	}
 
 	if (!os->service->chkput)
 		goto done;
 
-	ret = os->service->chkput(obex, obj);
+	ret = os->service->chkput(os, os->service_data);
 	switch (ret) {
 	case 0:
 		break;
-	case -EINVAL:
-		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
-				OBEX_RSP_BAD_REQUEST);
-		return FALSE;
 	case -EPERM:
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return FALSE;
+	case -EBADR:
+		OBEX_ObjectSetRsp(obj, OBEX_RSP_BAD_REQUEST,
+					OBEX_RSP_BAD_REQUEST);
+		return FALSE;
 	case -EAGAIN:
 		OBEX_SuspendRequest(obex, obj);
 		os->obj = obj;
@@ -708,6 +823,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_INTERNAL_SERVER_ERROR,
 				OBEX_RSP_INTERNAL_SERVER_ERROR);
 		return FALSE;
+
 	}
 
 	if (os->size == OBJECT_SIZE_DELETE || os->size == OBJECT_SIZE_UNKNOWN) {
@@ -723,6 +839,8 @@ done:
 
 static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 {
+	int err;
+
 	if (!os->service) {
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
 		return;
@@ -739,11 +857,13 @@ static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 			return;
 	}
 
-	os->service->put(obex, obj);
+	err = os->service->put(os, os->service_data);
+	if (err < 0)
+		os_set_response(obj, err);
 }
 
-static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
-					gint evt, gint cmd, gint rsp)
+static void obex_event(obex_t *obex, obex_object_t *obj, int mode,
+					int evt, int cmd, int rsp)
 {
 	struct obex_session *os;
 
@@ -754,17 +874,19 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	switch (evt) {
 	case OBEX_EV_PROGRESS:
 		if (os->service->progress)
-			os->service->progress(obex, obj);
+			os->service->progress(os, os->service_data);
 		break;
 	case OBEX_EV_ABORT:
 		os->aborted = TRUE;
 		if (os->service->reset)
-			os->service->reset(obex);
+			os->service->reset(os, os->service_data);
 		os_reset_session(os);
 		OBEX_ObjectSetRsp(obj, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
 		break;
 	case OBEX_EV_REQDONE:
 		switch (cmd) {
+		case OBEX_CMD_CONNECT:
+			break;
 		case OBEX_CMD_DISCONNECT:
 			OBEX_TransportDisconnect(obex);
 			break;
@@ -774,7 +896,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 		default:
 			os_session_mark_aborted(os);
 			if (os->service->reset)
-				os->service->reset(obex);
+				os->service->reset(os, os->service_data);
 			os_reset_session(os);
 			break;
 		}
@@ -826,8 +948,8 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 			break;
 		default:
 			debug("Unknown request: 0x%X", cmd);
-			OBEX_ObjectSetRsp(obj,
-				OBEX_RSP_NOT_IMPLEMENTED, OBEX_RSP_NOT_IMPLEMENTED);
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_NOT_IMPLEMENTED,
+						OBEX_RSP_NOT_IMPLEMENTED);
 			break;
 		}
 		break;
@@ -842,7 +964,8 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 		case -EAGAIN:
 			OBEX_SuspendRequest(obex, obj);
 			os->obj = obj;
-			os->driver->set_io_watch(os->object, handle_async_io, os);
+			os->driver->set_io_watch(os->object, handle_async_io,
+									os);
 			break;
 		default:
 			OBEX_ObjectSetRsp(obj,
@@ -855,14 +978,14 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	case OBEX_EV_STREAMEMPTY:
 		switch (obex_write_stream(os, obex, obj)) {
 		case -EPERM:
-			OBEX_ObjectSetRsp(obj,
-				OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
+			OBEX_ObjectSetRsp(obj, OBEX_RSP_FORBIDDEN,
+							OBEX_RSP_FORBIDDEN);
 			break;
 		case -EAGAIN:
 			OBEX_SuspendRequest(obex, obj);
 			os->obj = obj;
 			os->driver->set_io_watch(os->object, handle_async_io,
-							os);
+									os);
 			break;
 		default:
 			break;
@@ -882,15 +1005,7 @@ static void obex_event(obex_t *obex, obex_object_t *obj, gint mode,
 	}
 }
 
-void server_free(struct server *server)
-{
-	g_free(server->folder);
-	g_free(server->capability);
-	g_free(server->devnode);
-	g_free(server);
-}
-
-static void obex_handle_destroy(gpointer user_data)
+static void obex_handle_destroy(void *user_data)
 {
 	struct obex_session *os;
 	obex_t *obex = user_data;
@@ -898,79 +1013,48 @@ static void obex_handle_destroy(gpointer user_data)
 	os = OBEX_GetUserData(obex);
 
 	if (os->service && os->service->disconnect)
-		os->service->disconnect(obex);
+		os->service->disconnect(os, os->service_data);
 
 	obex_session_free(os);
 
 	OBEX_Cleanup(obex);
 }
 
-static gboolean tty_reinit(gpointer data)
-{
-	struct server *server = data;
-	GSList *l;
-	guint services = 0;
-
-	for (l = server->drivers; l; l = l->next) {
-		struct obex_service_driver *driver = l->data;
-
-		services |= driver->service;
-	}
-
-	tty_init(services, server->folder, server->capability,
-			server->symlinks, server->devnode);
-
-	server_free(server);
-
-	return FALSE;
-}
-
 static gboolean obex_handle_input(GIOChannel *io,
-				GIOCondition cond, gpointer user_data)
+				GIOCondition cond, void *user_data)
 {
 	obex_t *obex = user_data;
-	struct obex_session *os = OBEX_GetUserData(obex);
 
 	if (cond & (G_IO_HUP | G_IO_ERR | G_IO_NVAL)) {
 		error("obex_handle_input: poll event %s%s%s",
 				(cond & G_IO_HUP) ? "HUP " : "",
 				(cond & G_IO_ERR) ? "ERR " : "",
 				(cond & G_IO_NVAL) ? "NVAL " : "");
-		goto failed;
+		return FALSE;
 	}
 
 	if (OBEX_HandleInput(obex, 1) < 0) {
 		error("Handle input error");
-		goto failed;
+		return FALSE;
 	}
 
 	return TRUE;
-
-failed:
-	if (os->server->devnode) {
-		if (cond & G_IO_NVAL)
-			tty_closed();
-		else
-			g_idle_add(tty_reinit, os->server);
-	}
-
-	return FALSE;
 }
 
-gint obex_session_start(GIOChannel *io, struct server *server)
+int obex_session_start(GIOChannel *io, uint16_t tx_mtu, uint16_t rx_mtu,
+			struct obex_server *server)
 {
 	struct obex_session *os;
 	obex_t *obex;
-	gint ret, fd;
+	int ret, fd;
 
 	os = g_new0(struct obex_session, 1);
 
-	os->service = obex_service_driver_find(server->drivers, NULL, 0,
-						NULL, 0);
-	os->current_folder = g_strdup(server->folder);
+	os->service = obex_service_driver_find(server->drivers, NULL,
+							0, NULL, 0);
 	os->server = server;
-	os->rx_mtu = server->rx_mtu ? server->rx_mtu : DEFAULT_RX_MTU;
-	os->tx_mtu = server->tx_mtu ? server->tx_mtu : DEFAULT_TX_MTU;
+	os->rx_mtu = rx_mtu != 0 ? rx_mtu : DEFAULT_RX_MTU;
+	os->tx_mtu = tx_mtu != 0 ? tx_mtu : DEFAULT_TX_MTU;
 	os->size = OBJECT_SIZE_DELETE;
 
 	obex = OBEX_Init(OBEX_TRANS_FD, obex_event, 0);
@@ -1003,30 +1087,115 @@ gint obex_session_start(GIOChannel *io, struct server *server)
 	return 0;
 }
 
-gint obex_tty_session_stop(void)
+const char *obex_get_name(struct obex_session *os)
 {
-	GSList *l;
-
-	for (l = sessions; l != NULL; l = l->next) {
-		struct obex_session *os = l->data;
-
-		if (os->server->devnode && os->io)
-			g_io_channel_shutdown(os->io, TRUE, NULL);
-	}
-
-	return 0;
+	return os->name;
 }
 
-struct obex_session *obex_get_session(gpointer object)
+void obex_set_name(struct obex_session *os, const char *name)
 {
-	GSList *l;
+	g_free(os->name);
+	os->name = g_strdup(name);
+	debug("Name changed: %s", os->name);
+}
 
-	for (l = sessions; l; l = l->next) {
-		struct obex_session *os = l->data;
+ssize_t obex_get_size(struct obex_session *os)
+{
+	return os->size;
+}
 
-		if (os->object == object)
-			return os;
+const char *obex_get_type(struct obex_session *os)
+{
+	return os->type;
+}
+
+const char *obex_get_root_folder(struct obex_session *os)
+{
+	return os->server->folder;
+}
+
+uint16_t obex_get_service(struct obex_session *os)
+{
+	return os->service->service;
+}
+
+gboolean obex_get_symlinks(struct obex_session *os)
+{
+	return os->server->symlinks;
+}
+
+const char *obex_get_capability_path(struct obex_session *os)
+{
+	return os->server->capability;
+}
+
+gboolean obex_get_auto_accept(struct obex_session *os)
+{
+	return os->server->auto_accept;
+}
+
+int obex_remove(struct obex_session *os, const char *path)
+{
+	if (os->driver == NULL)
+		return -EINVAL;
+
+	return os->driver->remove(path);
+}
+
+/* TODO: find a way to do this for tty or fix syncevolution */
+char *obex_get_id(struct obex_session *os)
+{
+	GError *gerr = NULL;
+	char address[18];
+	uint8_t channel;
+
+	bt_io_get(os->io, BT_IO_RFCOMM, &gerr,
+			BT_IO_OPT_DEST, address,
+			BT_IO_OPT_CHANNEL, &channel,
+			BT_IO_OPT_INVALID);
+	if (gerr)
+		return NULL;
+
+	return g_strdup_printf("%s+%d", address, channel);
+}
+
+ssize_t obex_aparam_read(struct obex_session *os,
+		obex_object_t *obj, const uint8_t **buffer)
+{
+	obex_headerdata_t hd;
+	uint8_t hi;
+	uint32_t hlen;
+
+	OBEX_ObjectReParseHeaders(os->obex, obj);
+
+	while (OBEX_ObjectGetNextHeader(os->obex, obj, &hi, &hd, &hlen)) {
+		if (hi == OBEX_HDR_APPARAM) {
+			*buffer = hd.bs;
+			return hlen;
+		}
 	}
 
-	return NULL;
+	return -EBADR;
+}
+
+int obex_aparam_write(struct obex_session *os,
+		obex_object_t *obj, const uint8_t *data, unsigned int size)
+{
+	obex_headerdata_t hd;
+
+	hd.bs = data;
+
+	return OBEX_ObjectAddHeader(os->obex, obj,
+			OBEX_HDR_APPARAM, hd, size, 0);
+}
+
+int memcmp0(const void *a, const void *b, size_t n)
+{
+	if (a == NULL)
+		return -(a != b);
+
+	if (b == NULL)
+		return a != b;
+
+	return memcmp(a, b, n);
 }
